import re

from django.core.exceptions import ObjectDoesNotExist
from django.test import TestCase
from edc_constants.constants import UUID_PATTERN
from edc_identifier.models import IdentifierModel

from .models import SubjectScreening, SubjectScreeningWithEligibilitySimple


class TestScreening(TestCase):
    def test_model(self):
        for model_cls in [SubjectScreening, SubjectScreeningWithEligibilitySimple]:
            with self.subTest(model_cls=model_cls):
                obj = model_cls.objects.create(age_in_years=25)

                try:
                    IdentifierModel.objects.get(identifier=obj.screening_identifier)
                except ObjectDoesNotExist:
                    self.fail(f"Identifier unexpectedly not found. {obj.screening_identifier}")

                self.assertTrue(re.match(UUID_PATTERN, obj.subject_identifier))

                screening_identifier = obj.screening_identifier
                obj.save()
                obj.refresh_from_db()
                self.assertEqual(screening_identifier, obj.screening_identifier)



from typing import Optional, Type

from django.apps import apps as django_apps
from django.conf import settings
from django.db import models
from django.utils.html import format_html
from edc_constants.constants import NO, NORMAL, YES


def if_yes(value) -> bool:
    """Returns True if value is YES."""
    if value == NORMAL:
        return True
    return value == YES


def if_no(value) -> bool:
    """Returns True if value is NO."""
    return value == NO


def if_normal(value) -> bool:
    """Returns True if value is NORMAL."""
    return value == NORMAL


def get_subject_screening_app_label() -> Optional[str]:
    return get_subject_screening_model_name().split(".")[0]


def get_subject_screening_model_name() -> Optional[str]:
    return getattr(settings, "SUBJECT_SCREENING_MODEL", None)


def get_subject_screening_model_cls() -> Type[models.Model]:
    return django_apps.get_model(get_subject_screening_model_name())


def get_subject_screening(
    subject_screening_model: str, screening_identifier: str
) -> models.Model:
    """Returns the subject screening model instance or raises"""
    model_cls = django_apps.get_model(subject_screening_model)
    return model_cls.objects.get(screening_identifier=screening_identifier)


def format_reasons_ineligible(*str_values: str, delimiter=None) -> str:
    reasons = None
    delimiter = delimiter or "|"
    str_values = tuple(x for x in str_values if x is not None)
    if str_values:
        reasons = "".join(str_values)
        reasons = format_html(reasons.replace(delimiter, "<BR>"))
    return reasons


def eligibility_display_label(eligible) -> str:
    return "ELIGIBLE" if eligible else "not eligible"



from .age_evaluator import AgeEvaluator
from .gender_evaluator import GenderEvaluator
from .utils import (
    eligibility_display_label,
    format_reasons_ineligible,
    if_no,
    if_normal,
    if_yes,
)



from datetime import date

from dateutil.relativedelta import relativedelta
from django.db import models
from edc_utils.date import get_utcnow

from ..stubs import SubjectScreeningModelStub


class ScreeningMethodsModeMixin(models.Model):
    def __str__(self: SubjectScreeningModelStub):
        return f"{self.screening_identifier} {self.gender} {self.age_in_years}"

    def natural_key(self: SubjectScreeningModelStub):
        return tuple(self.screening_identifier)

    @staticmethod
    def get_search_slug_fields():
        return ["screening_identifier", "subject_identifier", "reference"]

    @property
    def estimated_dob(self: SubjectScreeningModelStub) -> date:
        return self.report_datetime.date() - relativedelta(years=self.age_in_years)

    class Meta:
        abstract = True



import re

from django.db import models
from edc_constants.constants import UUID_PATTERN
from edc_identifier import is_subject_identifier_or_raise
from edc_identifier.model_mixins import NonUniqueSubjectIdentifierModelMixin
from edc_search.model_mixins import SearchSlugModelMixin

from ..screening_identifier import ScreeningIdentifier
from ..stubs import SubjectScreeningModelStub


class ScreeningIdentifierModelMixin(
    NonUniqueSubjectIdentifierModelMixin, SearchSlugModelMixin, models.Model
):

    identifier_cls = ScreeningIdentifier
    screening_identifier_field_name: str = "screening_identifier"

    def save(self, *args, **kwargs):
        """Screening Identifier is always allocated."""
        if not self.id:
            setattr(
                self,
                self.screening_identifier_field_name,
                self.identifier_cls().identifier,
            )
        super().save(*args, **kwargs)  # type:ignore

    def update_subject_identifier_on_save(self: SubjectScreeningModelStub) -> str:
        """Overridden to not create a new study-allocated subject identifier on save.

        Instead just set subject_identifier to a UUID for uniqueness
        from subject_identifier_as_pk.

        The subject_identifier will be set upon consent.
        """
        if not self.subject_identifier:
            self.subject_identifier = self.subject_identifier_as_pk.hex
        else:
            # validate it is either a valid subject identifier or a
            # uuid/uuid.hex
            if not re.match(UUID_PATTERN, self.subject_identifier):
                is_subject_identifier_or_raise(self.subject_identifier, reference_obj=self)
        return self.subject_identifier

    def make_new_identifier(self) -> str:
        return self.subject_identifier_as_pk.hex

    class Meta:
        abstract = True



from edc_constants.constants import FEMALE, MALE


class GenderEvaluator:

    eligible_gender = [MALE, FEMALE]

    def __init__(self, gender=None, **kwargs) -> None:  # noqa
        self.eligible = False
        self.reasons_ineligible = None
        if gender in self.eligible_gender:
            self.eligible = True
        if not self.eligible:
            self.reasons_ineligible = []
            if gender not in [MALE, FEMALE]:
                self.reasons_ineligible.append(f"{gender} is an invalid gender.")



import re

from django import forms
from django.urls.base import reverse
from django.utils.safestring import mark_safe
from edc_constants.constants import UUID_PATTERN
from edc_dashboard.url_names import url_names


class AlreadyConsentedFormMixin:
    def clean(self: forms.ModelForm) -> dict:
        cleaned_data = super().clean()  # type:ignore
        r = re.compile(UUID_PATTERN)
        if (
            self.instance.id
            and self.instance.subject_identifier
            and not r.match(self.instance.subject_identifier)
        ):
            url_name = url_names.get("subject_dashboard_url")
            url = reverse(
                url_name,
                kwargs={"subject_identifier": self.instance.subject_identifier},
            )
            msg = mark_safe(
                "Not allowed. Subject has already consented. "
                f'See subject <A href="{url}">{self.instance.subject_identifier}</A>'
            )
            raise forms.ValidationError(msg)
        return cleaned_data
